////////////////////////////////////////////////////////////////////////
// vga_controller_800_60.vhd
////////////////////////////////////////////////////////////////////////
// Author : Ulrich Zoltn
//          Copyright 2006 Digilent, Inc.
////////////////////////////////////////////////////////////////////////
// Software version : Xilinx ISE 7.1.04i
//                    WebPack
// Device	        : 3s200ft256/4
////////////////////////////////////////////////////////////////////////
// This file contains the logic to generate the synchronization signals,
// horizontal and vertical pixel counter and video disable signal
// for the 800x600@60Hz resolution.
////////////////////////////////////////////////////////////////////////
//  Behavioral description
////////////////////////////////////////////////////////////////////////
// Please read the following article on the web regarding the
// vga video timings:
// http://www.epanorama.net/documents/pc/vga_timing.html

// This module generates the video synch pulses for the monitor to
// enter 800x600@60Hz resolution state. It also provides horizontal
// and vertical counters for the currently displayed pixel and a blank
// signal that is active when the pixel is not inside the visible screen
// and the color outputs should be reset to 0.

// timing diagram for the horizontal synch signal (HS)
// 0                         840    968          1056 (pixels)
// _________________________|//////|_________________
// timing diagram for the vertical synch signal (VS)
// 0                                  601    605  628 (lines)
// __________________________________|//////|________

// The blank signal is delayed one pixel clock period (25ns) from where
// the pixel leaves the visible screen, according to the counters, to
// account for the pixel pipeline delay. This delay happens because
// it takes time from when the counters indicate current pixel should
// be displayed to when the color data actually arrives at the monitor
// pins (memory read delays, synchronization delays).
////////////////////////////////////////////////////////////////////////
//  Port definitions
////////////////////////////////////////////////////////////////////////
// rst               / global reset signal
// pixel_clk         / input pin, from dcm_40MHz
//                   / the clock signal generated by a DCM that has
//                   / a frequency of 40MHz.
// HS                / output pin, to monitor
//                   / horizontal synch pulse
// VS                / output pin, to monitor
//                   / vertical synch pulse
// hcount            / output pin, 11 bits, to clients
//                   / horizontal count of the currently displayed
//                   / pixel (even if not in visible area)
// vcount            / output pin, 11 bits, to clients
//                   / vertical count of the currently active video
//                   / line (even if not in visible area)
// blank             / output pin, to clients
//                   / active when pixel is not in visible area.
////////////////////////////////////////////////////////////////////////
// Revision History:
// 09/18/2006(UlrichZ): created
// Port:
// Bjrn
////////////////////////////////////////////////////////////////////////

// the vga_controller_800_60 entity declaration
// read above for behavioral description and port definitions.
module vga_draw(CABCDEFGP,AN,RED,GRN,BLUE,HS,VS,clk,rst,button1,button2,button3,sw0);
`include "../pythonscript/params.v"
output [2:0] RED;
output [2:0] GRN;
output [1:0] BLUE;
output VS;
output HS;
output [7:0] CABCDEFGP;
output [3:0] AN;
input rst;
input clk;
input button1;
input button2;
input button3;
input sw0;

parameter VMAX  = 525;
parameter HMAX  = 800;

// total number of visible lines
parameter HLINES= 640;
// total number of visible lines
parameter VLINES=   480;

reg [2:0] RED;
reg [2:0] GRN;
reg [1:0] BLUE;

wire [2:0] REDG;
wire [2:0] GRNG;
wire [1:0] BLUEG;

wire [10:0] hcount;
wire [10:0] vcount;
wire blank;

wire [15:0] addr_ptr;
wire [3:0] bit_ptr;


reg [10:0] counter; 
reg [3:0] counter_seg;
reg [3:0] AN;
reg [1:0] ANUPDATE;
reg [19:0] counter_seg_update;
reg [3:0] counter_seg_thousands;
reg [3:0] counter_seg_hundreds;
reg [3:0] counter_seg_tens;
reg [7:0] CABCDEFGP;
reg [7:0] seg [0:9];

reg [9:0] vga_ram_hadr;
reg [9:0] vga_ram_vadr;
reg vga_ram_we;
wire [7:0] vga_ram_out;
reg [7:0] vga_ram_in;
reg [9:0]	pointerhadr;
reg start;
wire done;
reg [15:0] symbol;
reg [9:0]hpx;
reg [9:0]vpx;
wire [15:0] write_addr;
wire write_we;
wire [7:0] write_data;
reg [15:0] addr_in;
wire working;
wire clk40;

reg updatemenu1;
reg updatemenux;

parameter MAXCOUNT = 832000;
parameter	Timing1 =  208000;
parameter	Timing2 =  416000;
parameter	Timing3 =  624000;
parameter	Timing4 =  832000;


/*
reg [19:0] Timing1;
reg [19:0] Timing2;
reg [19:0] Timing3;
reg [19:0] Timing4;



*/
/*
always @(posedge button2 or posedge button3 or posedge rst) begin
	if(rst == 1'b1) begin
		DIM <= 19'd0;
	end else if (button2 == 1) begin
		DIM <= DIM+10000;
	end else if (button3 == 1) begin
		if(DIM>0) begin
			DIM <= DIM - 10000;
		end
	end
	Timing1 <=  208000-DIM;
	Timing2 <=  416000-DIM;
	Timing3 <=  624000-DIM;
	Timing4 <=  (MAXCOUNT-DIM);
end
*/



always @(posedge clk or posedge rst) begin
	if(rst == 1'b1) begin
		 seg[1] <= 8'b10011111;
		 seg[2] <= 8'b00100101;
		 seg[3] <= 8'b00001101;
		 seg[4] <= 8'b10011001;
		 seg[5] <= 8'b01001001;
		 seg[6] <= 8'b01000001;
		 seg[7] <= 8'b00011111;
		 seg[8] <= 8'b00000001;
		 seg[9] <= 8'b00001001;
		 seg[0] <= 8'b00000011;
		counter_seg_update <= 0;
		ANUPDATE <= 2'b0;
		CABCDEFGP <= 8'b0;
	end else begin
		if(counter_seg_update < Timing1) begin
			ANUPDATE <= 3;
			CABCDEFGP <= seg[counter_seg_thousands];
			counter_seg_update <= counter_seg_update+1;
		end else if(counter_seg_update < Timing2) begin
			ANUPDATE <= 2;
			CABCDEFGP <= seg[counter_seg_hundreds];
			counter_seg_update <= counter_seg_update+1;
		end else if(counter_seg_update < Timing3) begin
			ANUPDATE <= 1;
			CABCDEFGP <= seg[counter_seg_tens];
			counter_seg_update <= counter_seg_update+1;
		end else if(counter_seg_update < Timing4) begin
			ANUPDATE <= 0;
			CABCDEFGP <= seg[counter_seg];
			counter_seg_update <= counter_seg_update+1;
		end else begin
			counter_seg_update <= 0;
		end
	end
end


always @(posedge button1 or posedge rst) begin
	if(rst == 1'b1) begin
		counter_seg <= 0; 
		counter_seg_thousands <= 4'b0;
		counter_seg_hundreds <= 4'b0;
		counter_seg_tens <= 4'b0 ;
	end else begin
		if(counter_seg > 8) begin
			counter_seg <= 0;
			if(counter_seg_tens > 8 ) begin 
				counter_seg_tens <= 0;
				if(counter_seg_hundreds > 8 ) begin 
					counter_seg_hundreds <= 0;
					if(counter_seg_thousands > 8 ) begin 
						counter_seg_thousands <= 0;
					end else begin
						counter_seg_thousands <= counter_seg_thousands + 1;
					end
				end else begin
					counter_seg_hundreds <= counter_seg_hundreds+1;
				end
			end else begin
				counter_seg_tens <= counter_seg_tens+1;
			end
		end else begin
			counter_seg <= counter_seg + 1;
		end
	end
end



//refresh 7segment
always @* begin

case(ANUPDATE)
	0: begin
		AN[0] = 0;
		AN[1] = 1;
		AN[2] = 1;
		AN[3] = 1;
	end
	1: begin
		AN[0] = 1;
		AN[1] = 0;
		AN[2] = 1;
		AN[3] = 1;
	end
	2: begin
		AN[0] = 1;
		AN[1] = 1;
		AN[2] = 0;
		AN[3] = 1;
	end
	3: begin
		AN[0] = 1;
		AN[1] = 1;
		AN[2] = 1;
		AN[3] = 0;
	end
	default: begin
		AN[0] = 1;
		AN[1] = 1;
		AN[2] = 1;
		AN[3] = 1;
	end
	endcase
end


	
//VGA

//MUX TO SET COLOR OF IF BLANK=1;
always @* begin
	if(blank == 1'b1) begin
		RED <= 3'b000;
		GRN <= 3'b000;
		BLUE <= 2'b00;
	end else begin
		RED <= REDG;
		GRN <= GRNG;
		BLUE <= BLUEG;
	end
end

assign REDG[0] = vga_ram_out[bit_ptr[2:0]];
assign REDG[1] = vga_ram_out[bit_ptr[2:0]];//MemDB[0];
assign REDG[2] = vga_ram_out[bit_ptr[2:0]];
assign GRNG[0] = vga_ram_out[bit_ptr[2:0]];
assign GRNG[1] = vga_ram_out[bit_ptr[2:0]];
assign GRNG[2] = vga_ram_out[bit_ptr[2:0]];
assign BLUEG[0] = vga_ram_out[bit_ptr[2:0]];
assign BLUEG[1] = vga_ram_out[bit_ptr[2:0]];
parameter MENU1_MAX = 5;
parameter MENU2_MAX = 6;
parameter MENU3_MAX = 4;
reg button2_released;
reg button3_released;
reg [15:0]vpxhpx;
reg [3:0]menu1_ptr;
reg [3:0]menu2_ptr;
reg [3:0]menu3_ptr;

always @* begin
	if(rst == 1'b1) begin
		vga_ram_hadr <= 10'b0;
		vga_ram_vadr <= 10'b0;
		vga_ram_we <= 1'b0;
		vga_ram_in <= 8'b0;
	end else if(start==1 || working==1) begin
			vga_ram_we <= write_we;
			vga_ram_in <= write_data;
			//addr_in <= addr_ptr;
			addr_in <= write_addr;
	end else begin
			vga_ram_we <= 0;
			addr_in <= addr_ptr;
			if(blank == 0) begin
				vga_ram_vadr <= vcount[9:0];
				vga_ram_hadr <= hcount[9:0];
			end
	end
end

reg [31:0] button_counter;
reg button_timer_working;
reg button_timer_start;

parameter MAX_TIMER = 25000000;
always @(posedge clk or posedge rst) begin
	if (rst == 1) begin
		button_counter <= 0;
		button_timer_working <=32'b0;
	end else if(start == 1 && button_timer_working == 0) begin
		button_timer_working <= 1;
		button_counter <= button_counter+1;
	end else if(button_timer_working == 1 && (button_counter < MAX_TIMER)) begin
		button_counter <= button_counter+1;
	end else begin
		button_counter <= 32'b0;
		button_timer_working <=0;
	end
end

//Update menu pos thread
always @(posedge clk or posedge rst) begin
	if (rst==1) begin
		menu1_ptr <= 0;
		menu2_ptr <= 0;
		menu3_ptr <= 0;
		button3_released <= 1;
		button2_released <= 1;
		vpxhpx <= 16'b0;
		start <= 0;
		symbol <= 16'b0;
		updatemenu1 <= 0;
		updatemenux <= 0;
		button_timer_start <=0;
	end else if((working == 1) || (start == 1)) begin
		start <= 0;
	
	end else if(updatemenu1==1) begin
		//Then draw the new one
		if(menu1_ptr < MENU1_MAX) begin
			menu1_ptr <= menu1_ptr+1;
		end else begin
			menu1_ptr <= 0;
		end
		updatemenu1 <= 0;
		start <= 1;
		button3_released <= 0;
		symbol <= SYMBOL_PTR+17;
		if(menu1_ptr == 0) begin
			vpxhpx <= PX_MENU1_2;
		end else if (menu1_ptr == 1) begin
			vpxhpx <= PX_MENU1_3;
		end else if (menu1_ptr == 2) begin	
			vpxhpx <= PX_MENU1_4;
		end else if (menu1_ptr == 3) begin	
			vpxhpx <= PX_MENU1_5;
		end else if (menu1_ptr == 4) begin	
			vpxhpx <= PX_MENU1_1;
		end
		
		
	end else if(updatemenux == 1) begin
		//write new graphics
		start <= 1;
		button2_released <= 0;
		symbol <= SYMBOL_PTR+17;
		updatemenux <= 0;
		if(menu1_ptr == 0) begin
			//registers
			
			
		end else if(menu1_ptr == 1) begin
			//menu 1 is on program change
			if(menu3_ptr < MENU3_MAX) begin
				menu3_ptr <= menu3_ptr+1;
			end else begin
				menu3_ptr <= 0;
			end
			//Update menu GUI
			if(menu3_ptr == 0) begin
			//need a case for deleting old graphic
				vpxhpx <= PX_MENU3_2;
			end else if (menu3_ptr == 1) begin
				vpxhpx <= PX_MENU3_3;
			end else if (menu3_ptr == 2) begin	
				vpxhpx <= PX_MENU3_4;
			end else if (menu3_ptr == 3) begin	
				vpxhpx <= PX_MENU3_1;
			end
				
				
		end else if(menu1_ptr == 2) begin
			//Menu 1 is on clock change
			if(menu2_ptr < MENU2_MAX) begin
				menu2_ptr <= menu2_ptr+1;
			end else begin
				menu2_ptr <= 0;
			end
			//Update menu GUI
			if(menu2_ptr == 0) begin
				vpxhpx <= PX_MENU2_2;
			end else if (menu3_ptr == 1) begin
				vpxhpx <= PX_MENU2_3;
			end else if (menu2_ptr == 2) begin	
				vpxhpx <= PX_MENU2_4;
			end else if (menu2_ptr == 3) begin	
				vpxhpx <= PX_MENU2_5;
			end else if (menu2_ptr == 4) begin	
				vpxhpx <= PX_MENU2_6;
			end else if (menu2_ptr == 5) begin	
				vpxhpx <= PX_MENU2_1;
			end
		end
	end else if(button_timer_working==1 || button_timer_start == 1) begin
		button_timer_start <=0;
		
	end else if(button3==1 && button3_released == 1) begin
		//First delete old menu1 pointer GUI
		button_timer_start <=1;
		updatemenu1 <= 1;
		start <= 1;
		button3_released <= 0;
		symbol <= SYMBOL_PTR+18; // blank symbol
		if(menu1_ptr == 0) begin
		//need a case for deleting old graphic
			vpxhpx <= PX_MENU1_1;
		end else if (menu1_ptr == 1) begin
			vpxhpx <= PX_MENU1_2;
		end else if (menu1_ptr == 2) begin	
			vpxhpx <= PX_MENU1_3;
		end else if (menu1_ptr == 3) begin	
			vpxhpx <= PX_MENU1_4;
		end else if (menu1_ptr == 4) begin	
			vpxhpx <= PX_MENU1_5;
		end
	
	end else if(button2==1 && button2_released == 1) begin
		//First delete old graphics
		button_timer_start <=1;
		start <= 1;
		updatemenux <= 1;
		symbol <= SYMBOL_PTR+18;
		button2_released <= 0;
		if(menu1_ptr == 0) begin
			//registers
			
			
		end else if(menu1_ptr == 1) begin
			//menu 1 is on program change
			
			//Update Program menu pointer
			
			//Update menu GUI
			if(menu3_ptr == 0) begin
				vpxhpx <= PX_MENU3_1;
			end else if (menu3_ptr == 1) begin
				vpxhpx <= PX_MENU3_2;
			end else if (menu3_ptr == 2) begin	
				vpxhpx <= PX_MENU3_3;
			end else if (menu3_ptr == 3) begin	
				vpxhpx <= PX_MENU3_4;
			end
				
				
		end else if(menu1_ptr == 2) begin
			//Menu 1 is on clock change
			
			//Update menu pointer
			
			//Update menu GUI
			if(menu2_ptr == 0) begin
				vpxhpx <= PX_MENU2_1;
			end else if (menu2_ptr == 1) begin
				vpxhpx <= PX_MENU2_2;
			end else if (menu2_ptr == 2) begin	
				vpxhpx <= PX_MENU2_3;
			end else if (menu2_ptr == 3) begin	
				vpxhpx <= PX_MENU2_4;
			end else if (menu2_ptr == 4) begin	
				vpxhpx <= PX_MENU2_5;
			end else if (menu2_ptr == 5) begin	
				vpxhpx <= PX_MENU2_6;
			end
		end
		
	end else begin
		if(button3 == 0) begin
			button3_released <= 1;
		end
		if(button2 == 0) begin
			button2_released <= 1;
		end
	end
end

//Update GUI thread		
/*always@* begin
    if(rst == 1'b1) begin
		addr_in <= 0;
		vga_ram_hadr <= 10'b0;
		vga_ram_vadr <= 10'b0;
		vga_ram_we <= 1'b0;
		vga_ram_in <= 8'b0;
	end else begin
	if(sw0 == 1 && working == 0) begin
		symbol <= pointerhadr;
		vpx <= menu1_ptr;
		hpx <= menu1_ptr;
		addr_in <= write_addr;
		vga_ram_we <= write_we;
		vga_ram_in <= write_data;
	end else if(sw0 == 1 || working == 1) begin
			vpx <= menu1_ptr;
			hpx <= menu1_ptr;
			addr_in <= write_addr;
			if(working==0) begin
				start <= 1;
			end else begin
				start <= 0;
			end
			symbol <= pointerhadr;
			vga_ram_we <= write_we;
			vga_ram_in <= write_data;
		end else begin
			vga_ram_we <= 0;
			addr_in <= addr_ptr;
			if(blank ==0) begin
				vga_ram_vadr <= vcount[9:0];
				vga_ram_hadr <= hcount[9:0];
			end 
		end
	end
end		
	*/


vga_controller_800_60 VGACTRL1(	.rst(rst),
								.pixel_clk	(clk),
								.HS			(HS),
								.VS			(VS),
								.hcount		(hcount),
								.vcount		(vcount),
								.blank		(blank),
								.addr_ptr	(addr_ptr),
								.bit_ptr		(bit_ptr));
								
vga_ram VGARAM (.addr_ptr(addr_in[15:0]),.clk(clk), .in(vga_ram_in[7:0]), .we(vga_ram_we), .out(vga_ram_out),.rst(rst));
write_symbol WRITEMODULE(.start(start),.done(done),.working(working),.clk(clk),.symbol(symbol[15:0]),.vpxhpx (vpxhpx[15:0]),.mem_out(vga_ram_out),.mem_data(write_data),.addr(write_addr),.we(write_we),.rst(rst));
fredivider CLKDIV(.clk(clk),.rst(rst),.clk_out(clk40));
endmodule


module vga_controller_800_60(rst,pixel_clk,HS,VS,hcount,vcount,blank,addr_ptr,bit_ptr);
input rst;
input pixel_clk;
output HS;
output VS;
output [10:0] hcount;
output [10:0] vcount;
output [15:0] addr_ptr;
output [3:0] bit_ptr;
output blank;

////////////////////////////////////////////////////////////////////////
// CONSTANTS
////////////////////////////////////////////////////////////////////////

// maximum value for the horizontal pixel counter
parameter HMAX  = 800;
// maximum value for the vertical pixel counter
parameter VMAX  = 525;
// total number of visible columns
parameter HLINES= 640;
// value for the horizontal counter where front porch ends
parameter HFP   = 648;
// value for the horizontal counter where the synch pulse ends
parameter HSP   = 744;
// total number of visible lines
parameter VLINES=   480;
// value for the vertical counter where the front porch ends
parameter VFP   =   482;
// value for the vertical counter where the synch pulse ends
parameter VSP   =   484;
// polarity of the horizontal and vertical synch pulse
// only one polarity used, because for this resolution they coincide.
parameter SPP   = 0;

////////////////////////////////////////////////////////////////////////
// SIGNALS
////////////////////////////////////////////////////////////////////////

// horizontal and vertical counters
reg [10:0] hcount;
reg [10:0] vcount;
reg HS;
reg VS;
reg blank;

// active when inside visible screen area.
reg video_enable;
reg [15:0] addr_ptr;
reg [3:0] bit_ptr;

//RAM address managment
always @ (negedge pixel_clk or posedge rst) begin
	if(rst==1'b1) begin
		addr_ptr <= 16'b0;
		bit_ptr <= 4'd7;
	end else begin
		
		if(((hcount <= HLINES+1) && (vcount <= VLINES+1)) && (hcount > 1)) begin
		
			if(bit_ptr > 0) begin
				if(bit_ptr == 4'd2) begin
					addr_ptr <= addr_ptr +1'b1;
				end
					bit_ptr <= bit_ptr - 1'b1;
			end else begin
				bit_ptr <= 4'd7;
				//addr_ptr <= addr_ptr +1'b1;
			end
		end else if((hcount == HMAX) && (vcount == VMAX)) begin
				bit_ptr <= 4'd7;
				addr_ptr <= 16'b0;
		end
	end
end
	
	// blank is active when outside screen visible area
   // color output should be blacked (put on 0) when blank in active
   // blank is delayed one pixel clock period from the video_enable
   // signal to account for the pixel pipeline delay.
   always @ (posedge pixel_clk or posedge rst) begin
	if(rst==1'b1) begin
		blank <= 0;
	end else begin
		blank <= !video_enable;
	end
	end
   // increment horizontal counter at pixel_clk rate
   // until HMAX is reached, then reset and keep counting
   always @(posedge pixel_clk or posedge rst) begin
         if(rst == 1'b1) begin
            hcount <= 11'b0;
         end else if(hcount == HMAX) begin
            hcount <= 11'b0;
         end else begin
            hcount <= hcount + 1;
         end
   end

   // increment vertical counter when one line is finished
   // (horizontal counter reached HMAX)
   // until VMAX is reached, then reset and keep counting
  always @(posedge pixel_clk or posedge rst) begin
         if(rst == 1'b1) begin
            vcount <= 11'b0;
         end else if(hcount == HMAX) begin
            if(vcount == VMAX) begin
               vcount <= 11'b0;
            end else begin
               vcount <= vcount + 1;
            end
         end
  end

   // generate horizontal synch pulse
   // when horizontal counter is between where the
   // front porch ends and the synch pulse ends.
   // The HS is active (with polarity SPP) for a total of 128 pixels.
   always @(posedge pixel_clk or posedge rst) begin
	   if(rst == 1'b1) begin
            HS <= 1'b0;
       end else if(hcount >= HFP && hcount < HSP) begin
            HS <= SPP;
       end else begin
            HS <= !SPP;
        end
      
   end 

   // generate vertical synch pulse
   // when vertical counter is between where the
   // front porch ends and the synch pulse ends.
   // The VS is active (with polarity SPP) for a total of 4 video lines
   // = 4*HMAX = 4224 pixels.
   always @(posedge pixel_clk or posedge rst) begin
	   if(rst == 1'b1) begin
            VS <= 1'b0;
       end else if(vcount >= VFP && vcount < VSP) begin
            VS <= SPP;
       end else begin
            VS <= !SPP;
        end
      
   end 
   always @* begin
	video_enable <= (hcount < HLINES) && (vcount < VLINES);
   end
  endmodule
  /*
  module v_rams_07 (clk, we, a, di, do);
		input clk;
		input we;
		input [5:0] a;
		input [15:0] di;
		output [15:0] do;
		reg [15:0] ram [63:0];
		reg [5:0] read_a;
		always @(posedge clk) begin
			if (we)
				ram[a] <= di;
				read_a <= a;
		end
 assign do = ram[read_a];
endmodule
  */
  module vga_ram (addr_ptr, clk, in, we, out,rst);
  input [15:0] addr_ptr;
  input clk;
  input [7:0] in;
  input we,rst;
  output [7:0] out;
  


  //ext_mem VGAMEM2 (.OE(OE),.WE(mem_WE),.ADV(ADV),.CLK(mem_CLK),.UB(UB),.LB(LB),.CE(CE),.CRE(CRE),.ST_CE(ST_CE));

 // reg VGAROM [SIZE-1:0] [SIZE-1:0];
  //reg [18 : 0] addra;
  //integer i;
  //integer j;
	vga_mem640_8bit VGAMEM(
		.clka(clk),
		.wea(we),
		.addra(addr_ptr[15:0]),
		.dina(in[7:0]),
		.douta(out)
);
/*
  always @(posedge clk or posedge rst) begin
		if(rst==1) begin
			for(i=0; i<SIZE; i=i+1) begin
				for(j=0; j<SIZE; j=j+1) begin
					VGAROM[i][j] <= 1'b0;
				end
			end
		end else if(we==1) begin
			if((hadr<SIZE) && (vadr <SIZE))begin
				VGAROM[vadr][hadr] <= in;
			end
		end else begin
			if((hadr<SIZE) && (vadr <SIZE))begin
				out <= VGAROM[vadr][hadr];
			end else begin
				out <= 1'b1;
			end
		end
	end
	*/
  endmodule 
  
/*module addr_decoder(hpx,vpx,pxaddr,clk,rst);
	input [9:0]
	input clk,rst;
	output [15:0] pxaddr;
	wire [19:0] mulres;
	wire [5:0] div;
	//reg [15:0] pxaddr;
	assign div	= hpx[9:4];
	
	multiplier_tenbit multiplier (
	  .a(vpx [9:0]), // input [9 : 0] a
	  .b(10'd80), // input [9 : 0] b
	  .p(mulres) // output [19 : 0] p
	);
	assign pxaddr = div+mulres;
	/*always @(posedge clk or posedge rst) begin
	if(rst == 1) begin
		pxaddr <= 16'b0;
	end else begin
		pxaddr <= mulres[19:4];
	end
	end
	*/
//endmodule


module write_symbol(start,done,working,clk,symbol,vpxhpx,mem_out,mem_data,addr,we,rst);
	input start,clk,rst;
	input [15:0]symbol;
	input [15:0] vpxhpx;
	input [7:0] mem_out;
	output [7:0] mem_data;
	output [15:0] addr;
	output done,we,working;
	
	reg working;
	reg [2:0] counter;
	reg [15:0]addr;
	reg read;
	reg write;
	reg done;
	reg we; 
	reg workpipe;
	reg [7:0] mem_data;
	reg [7:0] readsymbol [0:7];
	integer i;
	//addr_decoder ADDRDEC (.hpx(hpx[9:0]),.vpx(vpx[9:0]),.pxaddr(pxaddr[15:0]),.clk(clk),.rst(rst));
	
	always @(posedge clk or posedge rst) begin
		if(rst == 1) begin
			addr <= 16'b0;
			for(i = 0; i<7; i=i+1) begin
				readsymbol[i] <= 8'b0;
			end
			counter <= 3'b0;
			read <= 1'b0;
			working <= 1'b0;
			write <= 1'b0;
			done <= 1'b1;
			we <= 1'b0;
			mem_data <= 8'b0;
			workpipe <= 1'b0;
		end else  begin
			if (start == 1'b1 || working == 1'b1) begin
			//FIRST READ BITS	
				if(working == 1'b0)begin
					working <=1'b1; 
					counter <= 3'b0;
					addr <= symbol[15:0];
					read <= 1'b1;
					done <= 1'b0; 
				end else if(workpipe == 0) begin
					workpipe <= 1;
					addr <= addr + 8'd80;
				end else if((working == 1'b1) && (read ==1'b1) &&  (counter < 3'd7)) begin
					addr <= addr + 8'd80;
					readsymbol[counter] <= mem_out;
					counter <= counter + 1 ;
				end else if ((working == 1'b1) && (read ==1'b1) && (counter ==3'd7)) begin
					readsymbol[counter] <= mem_out;
					write <= 1'b1;
					we <= 1'b1;
					read <= 1'b0;
					counter <= 3'd0;
					addr <= vpxhpx[15:0];
				end else if ((working == 1'b1) && (write ==1'b1) && (counter <3'd7)) begin
					addr <= addr + 8'd80;
					counter <= counter + 1;
					mem_data <= readsymbol[counter];
				end else if ((working == 1'b1) && (write ==1'b1) && (counter ==3'd7)) begin	
					mem_data <= readsymbol[counter];
					working <= 1'b0;
					write <= 1'b0;
					for(i = 0; i<8; i=i+1) begin
						readsymbol[i] <= 8'b0;
					end
					we <= 1'b0;
					counter <= 3'b0;
					addr <= symbol[15:0];
					done <= 1'b1;
					workpipe <= 0;
				end
		end
	end
end
				
endmodule
		
module fredivider(clk,rst,clk_out);
	input clk,rst;
	output reg clk_out;
	reg [15:0] counter;

always @(posedge clk or posedge rst) begin
	if(rst) begin
		counter<=16'd0;
		clk_out <= 1'b0;
	end else if(counter==16'd4) begin
		counter<=16'd0;
		clk_out <= ~clk_out;
	end else begin
	 counter<=counter+1;
	end
end

endmodule
  